""" Provides all the exceptions that may be raised.
"""


class UniqueConstraintError(Exception):
    """Raised when attempting to create more than one object with the same
       value on an attribute declared to be unique"""


class MultipleObjectsFound(Exception):
    """ Raised when a caller of a RelationshipManager
    expected a single object, but multiple were returned.
    """


class NoResultFound(Exception):
    """ Raised when a caller of a RelationshipManager
    expected at least one object, but none was found.
    """


class UnknownType(Exception):
    """ Raised when trying to deserialise a class that hasn't been
    registered
    """


class DeserialisationError(Exception):
    """ Raised when trying to deserialise a dict with no __type__ key """


class TypeAlreadyRegistered(Exception):
    pass



import pytest

from kaiso.exceptions import TypeAlreadyRegistered
from kaiso.types import Entity, PersistableMeta


def test_register_duplicate_fails():
    class Duplicate(Entity):
        pass

    with pytest.raises(TypeAlreadyRegistered):
        # same as declaring a class
        PersistableMeta('Duplicate', (Entity,), {})
        assert Duplicate is None



import pytest

from kaiso.types import PersistableMeta, Entity
from kaiso.exceptions import UnknownType


class Ente(Entity):
    pass


def test_type_ids():
    descr = PersistableMeta.get_descriptor(PersistableMeta)
    assert descr.type_id == 'PersistableMeta'

    descr = PersistableMeta.get_descriptor(Entity)
    assert descr.type_id == 'Entity'

    descr = PersistableMeta.get_descriptor(Ente)
    assert descr.type_id == 'Ente'


def test_unknown_type():
    with pytest.raises(UnknownType):
        PersistableMeta.get_descriptor(UnknownType)



import pytest

from kaiso.attributes import String
from kaiso.types import Entity


@pytest.mark.usefixtures('storage')
def test_save_dynamic_type(storage):

    attrs = {'id': String(unique=True)}
    Foobar = storage.create_type('Foobar', (Entity,), attrs)

    storage.save(Foobar)

    rows = storage.query('START n=node:persistablemeta(id="Foobar") RETURN n')
    (result,) = next(rows)

    assert result is Foobar


@pytest.mark.usefixtures('storage')
def test_save_dynamic_typed_obj(storage):

    attrs = {'id': String(unique=True)}
    Foobar = storage.create_type('Foobar', (Entity,), attrs)

    foo = Foobar(id='spam')
    storage.save(foo)

    rows = storage.query('START n=node:foobar(id="spam") RETURN n')
    (result,) = next(rows)

    assert result.id == foo.id


@pytest.mark.usefixtures('storage')
def test_add_attr_to_type(storage):
    Foobar = storage.create_type('Foobar', (Entity,), {})
    storage.save(Foobar)

    Foobar.ham = String(default='eggs')
    storage.save(Foobar)

    rows = storage.query(
        'START n=node:persistablemeta(id="Foobar") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 1


@pytest.mark.usefixtures('storage')
def test_remove_attr_from_type(storage):
    attrs = {'ham': String()}
    Foobar = storage.create_type('Foobar', (Entity,), attrs)
    storage.save(Foobar)

    del Foobar.ham
    storage.save(Foobar)

    rows = storage.query(
        'START n=node:persistablemeta(id="Foobar") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 0


@pytest.mark.usefixtures('storage')
def test_removing_attr_from_declared_type_does_not_remove_it(storage):

    # the use case:
    # developer defines a partial type in code
    # the type Ham gets updated in the DB at runtime
    # then the code changes, removing an attribute from the class
    # the type definition in the DB is not affected, nor
    # the data gets back from the DB

    class Ham(Entity):
        egg = String

    storage.save(Ham)

    attrs = {'egg': String(), 'spam': String()}
    DynHam = storage.create_type('Ham', (Entity,), attrs)
    storage.save(DynHam)

    del Ham.egg
    storage.save(Ham)

    rows = storage.query(
        'START n=node:persistablemeta(id="Ham") '
        'MATCH n <-[:DECLAREDON]- attr '
        'RETURN count(attr)')
    (count,) = next(rows)
    assert count == 2


@pytest.mark.usefixtures('storage')
def test_load_dynamic_types(storage):
    Animal = storage.create_type('Animal', (Entity,), {'id': String()})
    Horse = storage.create_type('Horse', (Animal,), {'hoof': String()})
    Duck = storage.create_type('Duck', (Animal,), {'beek': String()})
    Beaver = storage.create_type('Beaver', (Animal,), {'tail': String()})
    Platypus = storage.create_type(
        'Platypus', (Duck, Beaver), {'egg': String()})

    storage.save(Horse)
    storage.save(Platypus)

    # this is the same as creating a new storage
    storage.initialize()

    rows = storage.query(
        '''
        START ts=node:typesystem(id="TypeSystem")
        MATCH p = (ts -[:DEFINES]-> () <-[:ISA*0..]- tpe),
            tpe <-[:DECLAREDON*0..]- attr,
            tpe -[:ISA*0..1]-> base
        RETURN tpe.id,  length(p) AS level,
            filter(b_id in collect(distinct base.id): b_id <> tpe.id),
            collect(distinct attr.name?)
        ORDER BY level
        ''')
    result = list(rows)

    assert result == [
        ('Entity', 1, [], []),
        ('Animal', 2, ['Entity'], ['id']),
        ('Duck', 3, ['Animal'], ['beek']),
        ('Beaver', 3, ['Animal'], ['tail']),
        ('Horse', 3, ['Animal'], ['hoof']),
        ('Platypus', 4, ['Duck', 'Beaver'], ['egg']),
    ]


@pytest.mark.usefixtures('storage')
def test_add_attr_to_type_via_2nd_storage(storage):
    attrs = {'id': String(unique=True)}
    Shrub = storage.create_type('Shrub', (Entity,), attrs)

    shrub = Shrub(id='spam')
    storage.save(shrub)

    # this is the same as creating a new storage using the same URL
    storage.initialize()

    (Shrub,) = next(storage.query(
        'START cls=node:persistablemeta(id="Shrub") RETURN cls'))
    Shrub.ham = String(default='eggs')
    storage.save(Shrub)

    # we want to query from an independent storage
    storage.initialize()
    rows = storage.query('START n=node:shrub(id="spam") RETURN n')
    (result,) = next(rows)

    assert result.ham == 'eggs'



import os
import subprocess
import shutil

from mock import patch
import pytest

import kaiso.connection
from kaiso.connection import get_connection, TempConnectionError


@pytest.mark.slow
class TestTempConnectionProcesses():
    """ Test spinning up temporary neo4j processes.

    We mock out ``py2neo.neo4j.GraphDatabaseService`` because:
        a) it's a library and we're not testing its behaviour directly
        b) the first thing GraphDatabaseService does is generate a
           GET request (the same one we do in ``get_connection``)
           whch sometimes results in a SocketError. This is possibly
           a bug in neo4j, but it shouldn't influence these tests.
    """
    temp_data_dir = '/tmp/foo'

    def teardown_method(self, method):
        """ Kill process and remove temp_data_dir after every test.
        """
        for key, proc in kaiso.connection._temporary_databases.items():
            proc.terminate()
            proc.wait()
            del kaiso.connection._temporary_databases[key]
        if os.path.exists(self.temp_data_dir):
            shutil.rmtree(self.temp_data_dir)

    def test_temp_connection_defaults(self):
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://')
            assert conn == "http://localhost:7475/db/data/"

    def test_temp_connection_custom_port(self):
        port = "7777"
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}'.format(port))
            assert conn == "http://localhost:{}/db/data/".format(port)

    def test_temp_connection_custom_data_dir(self):
        data_dir = self.temp_data_dir
        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}'.format(data_dir))
            assert conn == "http://localhost:7475/db/data/"
            assert os.path.exists(data_dir)

    def test_temp_connection_custom(self):
        port = "7777"
        data_dir = self.temp_data_dir

        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn = get_connection('temp://{}{}'.format(port, data_dir))
            assert conn == "http://localhost:{}/db/data/".format(port)
            assert os.path.exists(data_dir)

    def test_multiple_temp_connections(self):

        with patch('py2neo.neo4j.GraphDatabaseService', lambda uri: uri):
            conn1 = get_connection('temp://')
            with patch.object(kaiso.connection, 'get_neo4j_info') as get_info:
                conn2 = get_connection('temp://')
                assert conn1 == conn2
                assert not get_info.called  # we should be reusing existing db

    def test_temp_connection_timeout(self):
        with patch.object(kaiso.connection, 'TIMEOUT', 0):
            with pytest.raises(TempConnectionError):
                get_connection('temp://8888')


def test_temp_connection_no_neo4j_info():
    with patch.object(subprocess, 'check_output', lambda _: None):
        with pytest.raises(TempConnectionError):
            get_connection('temp://8888')



from inspect import getmembers, getmro

from kaiso.exceptions import UnknownType, TypeAlreadyRegistered


def get_declaring_class(cls, attr_name):
    """ Returns the class in the type heirarchy of ``cls`` that defined
        an attribute with name ``attr_name``.
    """
    declaring_class = None
    declared_attr = None

    # Start at the top of the hierarchy and determine which of the MRO have
    # the attribute. Return the lowest class that defines (or overloads) the
    # attribute.
    for base in reversed(getmro(cls)):
        attr = getattr(base, attr_name, False)
        if attr and declared_attr is not attr:
            declaring_class = base
            declared_attr = attr

    return declaring_class


def get_index_name(cls):
    """ Returns a cypher index name for a class.

    Args:
        cls: The class to generate an index for.

    Returns:
        An index name.
    """
    if issubclass(cls, PersistableMeta):
        return cls.index_name
    else:
        return cls.__name__.lower()


def get_index_entries(obj):
    """ Returns tuples of (index-name, key, value) for a persistable object.

    It can be used to create queryies with index lookups.

    Args:
        obj: A persistable object.

    Returns:
        Generator yielding tuples (index_name, key, value)
    """

    cls = type(obj)

    if issubclass(cls, PersistableMeta):
        meta_cls = cls
    else:
        meta_cls = type(cls)

    return meta_cls.get_index_entries(obj)


def is_indexable(cls):
    """ Returns True if the class ``cls`` has any indexable attributes.
    """

    descr = Descriptor(cls)
    for _, attr in descr.attributes.items():
        if attr.unique:
            return True

    return False


class Descriptor(object):
    """ Provides information about the types of persistable objects.

    Its main purpose is to provide type names and attribute information of
    persistable types(classes).
    """
    def __init__(self, cls):
        self.cls = cls

    @property
    def type_id(self):
        cls = self.cls

        if issubclass(cls, PersistableMeta):
            return cls.type_id
        else:
            return cls.__name__

    @property
    def relationships(self):
        from kaiso.attributes.bases import _is_relationship_reference
        relationships = dict(getmembers(
            self.cls, _is_relationship_reference
        ))
        return relationships

    @property
    def attributes(self):
        attributes = dict(getmembers(self.cls, _is_attribute))
        return attributes

    @property
    def declared_attributes(self):
        declared = {}
        for name, attr in getmembers(self.cls, _is_attribute):
            if get_declaring_class(self.cls, name) == self.cls:
                declared[name] = attr

        return declared


class Persistable(object):
    """ The base of all persistable objects.

    Any object stored in the db must inherit from this class.
    Any object having Persistable as it's base are considered persistable.
    """


class DiscriptorType(type):
    def __init__(cls, name, bases, dct):  # pylint:disable-msg=W0613
        super(DiscriptorType, cls).__init__(cls)
        cls.descriptors = {}
        cls.register(cls)


class AttributeBase(object):
    name = None


class PersistableMeta(type, Persistable):
    __metaclass__ = DiscriptorType

    index_name = 'persistablemeta'
    type_id = 'PersistableMeta'

    def __new__(mcs, name, bases, dct):
        for attr_name, attr in dct.items():
            if isinstance(attr, AttributeBase):
                attr.name = attr_name

        cls = super(PersistableMeta, mcs).__new__(mcs, name, bases, dct)
        mcs.register(cls)
        return cls

    @classmethod
    def register(mcs, cls):
        name = cls.__name__
        if name in mcs.descriptors:
            raise TypeAlreadyRegistered(cls)
        mcs.descriptors[name] = Descriptor(cls)

    @classmethod
    def get_class_by_id(mcs, cls_id):
        try:
            if mcs is not PersistableMeta and issubclass(mcs, PersistableMeta):
                return PersistableMeta.get_class_by_id(cls_id)
        except UnknownType:
            pass

        return mcs.get_descriptor_by_id(cls_id).cls

    @classmethod
    def get_descriptor(mcs, cls):
        name = cls.__name__
        return mcs.get_descriptor_by_id(name)

    @classmethod
    def get_descriptor_by_id(mcs, cls_id):
        try:
            return mcs.descriptors[cls_id]
        except KeyError:
            if mcs is not PersistableMeta and issubclass(mcs, PersistableMeta):
                return PersistableMeta.get_descriptor_by_id(cls_id)

            raise UnknownType('Unknown type "{}"'.format(cls_id))

    @classmethod
    def get_index_entries(mcs, obj):
        if isinstance(obj, type):
            yield (mcs.index_name, 'id', obj.__name__)
        else:
            obj_type = type(obj)
            descr = mcs.get_descriptor(obj_type)

            for name, attr in descr.attributes.items():
                if attr.unique:
                    declaring_class = get_declaring_class(descr.cls, name)

                    index_name = get_index_name(declaring_class)
                    key = name
                    value = attr.to_db(getattr(obj, name))

                    if value is not None:
                        yield (index_name, key, value)


PersistableMeta.register(type)
PersistableMeta.register(object)


def _is_attribute(obj):
    return isinstance(obj, Attribute)


class Attribute(AttributeBase):
    __metaclass__ = PersistableMeta

    unique = None
    required = None

    def __init__(self, unique=False, required=False):
        self.unique = unique
        self.required = required

    @staticmethod
    def to_python(value):
        return value

    @staticmethod
    def to_db(value):
        return value


class DefaultableAttribute(Attribute):
    default = None

    def __init__(self, default=None, **kwargs):
        super(DefaultableAttribute, self).__init__(**kwargs)
        self.default = default


class AttributedBase(Persistable):
    """ The base class for objects that can have Attributes.

    Sets default values during instance creation and applies kwargs
    passed to __init__.
    """
    __metaclass__ = PersistableMeta

    def __new__(cls, *args, **kwargs):

        obj = super(AttributedBase, cls).__new__(cls, *args, **kwargs)

        descriptor = type(cls).get_descriptor(cls)

        for name, attr in descriptor.attributes.items():
            setattr(obj, name, attr.default)

        return obj

    def __init__(self, *args, **kwargs):
        super(AttributedBase, self).__init__(*args, **kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)


class Entity(AttributedBase):
    def __getattribute__(self, name):
        cls = type(self)
        descriptor = type(cls).get_descriptor(cls)
        try:
            rel_reference = descriptor.relationships[name]
        except KeyError:
            return object.__getattribute__(self, name)
        else:
            return rel_reference.get_manager(self)


class Relationship(AttributedBase):
    def __init__(self, start=None, end=None, **kwargs):
        super(Relationship, self).__init__(**kwargs)

        self.start = start
        self.end = end



from kaiso.attributes.bases import get_attibute_for_type
from kaiso.exceptions import DeserialisationError
from kaiso.iter_helpers import unique
from kaiso.relationships import InstanceOf, IsA
from kaiso.types import (
    Attribute, DefaultableAttribute, Descriptor, PersistableMeta)


def get_changes(old, new):
    """Return a changes dictionary containing the key/values in new that are
       different from old. Any key in old that is not in new will have a None
       value in the resulting dictionary
    """
    changes = {}

    # check for any keys that have changed, put their new value in
    for key, value in new.items():
        if old.get(key) != value:
            changes[key] = value

    for key in old.keys():
        if key not in new:
            raise KeyError('missing key: {}'.format(key))

    return changes


@unique
def get_type_relationships(obj):
    """ Generates a list of the type relationships of an object.
    e.g.
        get_type_relationships(Entity())

        (object, InstanceOf, type),
        (type, IsA, object),
        (type, InstanceOf, type),
        (PersistableMeta, IsA, type),
        (PersistableMeta, InstanceOf, type),
        (Entity, IsA, object),
        (Entity, InstanceOf, PersistableMeta),
        (<Entity object>, InstanceOf, Entity)

    Args:
        obj:    An object to generate the type relationships for.

    Returns:
        A generator, generating tuples
            (object, relatsionship type, related obj)
    """
    obj_type = type(obj)

    if obj_type is not type:
        for item in get_type_relationships(obj_type):
            yield item

    if isinstance(obj, type):
        for base in obj.__bases__:
            for item in get_type_relationships(base):
                yield item
            yield obj, IsA, base

    yield obj, InstanceOf, obj_type


def object_to_db_value(obj):
    try:
        attr_cls = get_attibute_for_type(type(obj))
    except KeyError:
        return obj
    else:
        return attr_cls.to_db(obj)


def dict_to_db_values_dict(data):
    return dict((k, object_to_db_value(v)) for k, v in data.items())


def object_to_dict(obj, type_registry, include_none=True):
    """ Converts a persistable object to a dict.

    The generated dict will contain a __type__ key, for which the value
    will be the type_id as given by the descriptor for type(obj).

    If the object is a class, dict will contain an id-key with the value
    being the type_id given by the descriptor for the object.

    For any other object all the attributes as given by the object's
    type descriptpr will be added to the dict and encoded as required.

    Args:
        obj: A persistable  object.

    Returns:
        Dictionary with attributes encoded in basic types
        and type information for deserialization.
        e.g.
        {
            '__type__': 'Entity',
            'attr1' : 1234
        }
    """
    obj_type = type(obj)

    properties = {
        '__type__': Descriptor(obj_type).type_id,
    }

    if isinstance(obj, type):
        properties['id'] = Descriptor(obj).type_id

    elif isinstance(obj, Attribute):
        # TODO: move logic to handle Attribute attrs into descriptor
        #       and let this code just treat them like any other persistable
        if obj.name is not None or include_none:
            properties['name'] = obj.name

        properties['unique'] = obj.unique
        properties['required'] = obj.required

        if isinstance(obj, DefaultableAttribute):
            if obj.default is not None or include_none:
                properties['default'] = obj.default

    else:
        descr = type_registry.get_descriptor(obj_type)

        for name, attr in descr.attributes.items():
            try:
                value = attr.to_db(getattr(obj, name))
            except AttributeError:
                # if we are dealing with an extended type, we may not
                # have the attribute set on the instance
                if isinstance(attr, DefaultableAttribute):
                    value = attr.default
                else:
                    value = None

            if value is not None or include_none:
                properties[name] = value

    return properties


def dict_to_object(properties, type_registry):
    """ Converts a dict into a persistable object.

    The properties dict needs at least a __type__ key containing the name
    of any registered class.
    The type key defines the type of the object to return.

    If the registered class for the __type__ is a meta-class,
    i.e. a subclass of <type>, a name key is assumed to be present and
    the registered class idendified by it's value is returned.

    If the registered class for the __type__ is standard class,
    i.e. an instance of <type>, and object of that class will be created
    with attributes as defined by the remaining key-value pairs.

    Args:
        properties: A dict like object.

    Returns:
        A persistable object.
    """

    try:
        type_id = properties['__type__']
    except KeyError:
        raise DeserialisationError(
            'properties "{}" missing __type__ key'.format(properties))

    if type_id == Descriptor(PersistableMeta).type_id:
        # we are looking at a class object
        cls_id = properties['id']
    else:
        # we are looking at an instance object
        cls_id = type_id

    cls = type_registry.get_class_by_id(cls_id)

    if cls_id != type_id:
        return cls
    else:
        obj = cls.__new__(cls)

        if isinstance(obj, Attribute):
            for attr_name, value in properties.iteritems():
                setattr(obj, attr_name, value)
        else:
            descr = type_registry.get_descriptor_by_id(cls_id)

            for attr_name, attr in descr.attributes.items():
                try:
                    value = properties[attr_name]
                except KeyError:
                    pass
                else:
                    value = attr.to_python(value)
                    setattr(obj, attr_name, value)

    return obj



import pytest

from kaiso.exceptions import DeserialisationError, UnknownType
from kaiso.attributes import String, Uuid
from kaiso.relationships import Relationship, InstanceOf, IsA
from kaiso.serialize import (
    get_type_relationships, get_changes,
    object_to_dict, dict_to_object, dict_to_db_values_dict)
from kaiso.types import (
    Persistable, DiscriptorType, PersistableMeta,
    Entity, AttributedBase, Attribute)


class Foo(Entity):
    pass


class Bar(Attribute):
    pass


class Spam(Entity):
    id = String()
    ham = String(default='eggs')


def test_classes_to_dict():
    dct = object_to_dict(Entity, PersistableMeta)
    assert dct == {'__type__': 'PersistableMeta', 'id': 'Entity'}

    obj = dict_to_object(dct, PersistableMeta)
    assert obj is Entity

    dct = object_to_dict(Foo, PersistableMeta)
    assert dct == {'__type__': 'PersistableMeta', 'id': 'Foo'}

    obj = dict_to_object(dct, PersistableMeta)
    assert obj is Foo


def test_objects():
    dct = object_to_dict(Entity(), PersistableMeta)
    assert dct == {'__type__': 'Entity'}

    obj = dict_to_object(dct, PersistableMeta)
    assert isinstance(obj, Entity)

    dct = object_to_dict(Foo(), PersistableMeta)
    assert dct == {'__type__': 'Foo'}

    obj = dict_to_object(dct, PersistableMeta)
    assert isinstance(obj, Foo)


def test_attribute():
    """ Verify (de)serialization of Attributes.
    """
    # Attribute dicts always contain both ``unique`` and ``required`` keys.
    attr = Bar(unique=True)
    dct = object_to_dict(attr, PersistableMeta)
    assert dct == {
        '__type__': 'Bar', 'unique': True, 'required': False, 'name': None}

    # Attribute objects have default values of ``None`` for ``unique``
    # and ``required``.
    obj = dict_to_object(
        {'__type__': 'Bar', 'unique': True, 'name': None}, PersistableMeta)
    assert isinstance(obj, Bar)
    assert obj.unique is True
    assert obj.required is None
    assert obj.name is None


def test_relationship():
    dct = object_to_dict(Relationship(None, None), PersistableMeta)
    assert dct == {'__type__': 'Relationship'}

    obj = dict_to_object(dct, PersistableMeta)
    assert isinstance(obj, Relationship)


def test_obj_with_attrs():
    spam = Spam(id=None)

    dct = object_to_dict(spam, PersistableMeta, include_none=False)
    assert dct == {'__type__': 'Spam', 'ham': 'eggs'}

    dct = object_to_dict(spam, PersistableMeta, include_none=True)
    assert dct == {'__type__': 'Spam', 'id': None, 'ham': 'eggs'}

    obj = dict_to_object(dct, PersistableMeta)
    assert obj.id == spam.id
    assert obj.ham == spam.ham

    dct.pop('ham')  # removing an attr with defaults
    obj = dict_to_object(dct, PersistableMeta)
    assert obj.id == spam.id
    assert obj.ham == spam.ham


def test_dynamic_type():
    DynamicType = type('DynamicType', (PersistableMeta,), {})

    Foobar = DynamicType('Foobar', (Entity,), {})

    dct = object_to_dict(Foobar, PersistableMeta)
    assert dct == {'__type__': 'PersistableMeta', 'id': 'Foobar'}

    # since Foobar is only registered with DynamicType
    with pytest.raises(UnknownType):
        obj = dict_to_object(dct, PersistableMeta)

    obj = dict_to_object(dct, DynamicType)
    assert obj is Foobar


def test_dynamic_typed_object():
    DynamicType = type('DynamicType', (PersistableMeta,), {})

    attrs = {'id': String()}
    Foobar = DynamicType('Foobar', (Entity,), attrs)

    foo = Foobar(id='spam')
    dct = object_to_dict(foo, DynamicType)

    assert dct == {'__type__': 'Foobar', 'id': 'spam'}

    obj = dict_to_object(dct, DynamicType)
    assert isinstance(obj, Foobar)
    assert obj.id == foo.id


def test_extended_declared_type_using_dynamic_type():
    DynamicType = type('DynamicType', (PersistableMeta,), {})

    attrs = {'id': Uuid()}
    DynEntity = DynamicType('Entity', (AttributedBase,), attrs)

    foo = DynEntity()
    dct = object_to_dict(foo, DynamicType)

    assert dct == {'__type__': 'Entity', 'id': str(foo.id)}

    obj = dict_to_object(dct, DynamicType)
    assert isinstance(obj, Entity)
    assert not isinstance(obj, DynEntity)
    assert obj.id == foo.id


def test_extended_declared_type_using_declared_type():
    DynamicType = type('DynamicType', (PersistableMeta,), {})

    attrs = {'id': String()}
    DynamicType('Entity', (AttributedBase,), attrs)

    dct = object_to_dict(Entity(), DynamicType)
    assert dct == {'__type__': 'Entity', 'id': None}


def test_extended_declared_type_with_default_using_declared_type():
    DynamicType = type('DynamicType', (PersistableMeta,), {})

    attrs = {
        'id': String(default='foobar'),
        'spam': Uuid()
    }
    DynamicType('Entity', (AttributedBase,), attrs)

    dct = object_to_dict(Entity(), DynamicType)
    assert dct == {'__type__': 'Entity', 'id': 'foobar', 'spam': None}


def test_changes():
    a = {'a': 12, 'b': '34'}
    b = {'a': 1, 'b': '34'}

    changes = get_changes(a, b)
    assert changes == {'a': 1}

    with pytest.raises(KeyError):
        get_changes(a, {})


def test_attribute_types():
    uid = Uuid().default
    dct = dict_to_db_values_dict({'foo': uid, 'bar': 123})

    assert dct == {'foo': str(uid), 'bar': 123}


def test_missing_info():
    with pytest.raises(DeserialisationError):
        dict_to_object({}, PersistableMeta)


def test_IsA_and_InstanceOf_type_relationships():
    """ Testing type hierarchy creation.

    We don't want to test all the types in their relationships,
    but only the set of types which are persisted with IsA and InstanceOf
    relationships.
    """
    result = list(get_type_relationships(Entity))

    assert result == [
        (object, InstanceOf, type),
        (type, IsA, object),
        (type, InstanceOf, type),
        (DiscriptorType, IsA, type),
        (DiscriptorType, InstanceOf, type),
        (PersistableMeta, IsA, type),
        (Persistable, IsA, object),
        (Persistable, InstanceOf, type),
        (PersistableMeta, IsA, Persistable),
        (PersistableMeta, InstanceOf, DiscriptorType),
        (AttributedBase, IsA, Persistable),
        (AttributedBase, InstanceOf, PersistableMeta),
        (Entity, IsA, AttributedBase),
        (Entity, InstanceOf, PersistableMeta),
    ]

    pers = Entity()
    result = list(get_type_relationships(pers))

    assert result == [
        (object, InstanceOf, type),
        (type, IsA, object),
        (type, InstanceOf, type),
        (DiscriptorType, IsA, type),
        (DiscriptorType, InstanceOf, type),
        (PersistableMeta, IsA, type),
        (Persistable, IsA, object),
        (Persistable, InstanceOf, type),
        (PersistableMeta, IsA, Persistable),
        (PersistableMeta, InstanceOf, DiscriptorType),
        (AttributedBase, IsA, Persistable),
        (AttributedBase, InstanceOf, PersistableMeta),
        (Entity, IsA, AttributedBase),
        (Entity, InstanceOf, PersistableMeta),
        (pers, InstanceOf, Entity),
    ]

    foo = Foo()
    result = list(get_type_relationships(foo))

    assert result == [
        (object, InstanceOf, type),
        (type, IsA, object),
        (type, InstanceOf, type),
        (DiscriptorType, IsA, type),
        (DiscriptorType, InstanceOf, type),
        (PersistableMeta, IsA, type),
        (Persistable, IsA, object),
        (Persistable, InstanceOf, type),
        (PersistableMeta, IsA, Persistable),
        (PersistableMeta, InstanceOf, DiscriptorType),
        (AttributedBase, IsA, Persistable),
        (AttributedBase, InstanceOf, PersistableMeta),
        (Entity, IsA, AttributedBase),
        (Entity, InstanceOf, PersistableMeta),
        (Foo, IsA, Entity),
        (Foo, InstanceOf, PersistableMeta),
        (foo, InstanceOf, Foo),
    ]



""" Provides a connection factory for connecting to existing or
temporary neo4j instances.

For temporary instances it requires the neo4j command
to be available on the path.
"""
import os
import re
import time
import atexit
import logging
import requests
import tempfile
import subprocess

from py2neo import neo4j

logger = logging.getLogger(__name__)

_temporary_databases = {}

TIMEOUT = 30  # seconds


class TempConnectionError(Exception):
    pass


def get_neo4j_info():
    """ Gets runtime information from the neo4j command.

    Returns:
        A dict.
    """
    default_cmd = os.pathsep.join([
        'neo4j',
        '/etc/init.d/neo4j-service'
    ])
    neo4j_cmds = os.environ.get('NEO4J_CMD', default_cmd).split(os.pathsep)
    output = None
    for cmd in neo4j_cmds:
        try:
            output = subprocess.check_output([cmd, 'info'])
        except OSError:
            pass

    if not output:
        raise TempConnectionError('Cannot determine neo4j info. Is the '
                                  'NEO4J_CMD environment varaible set '
                                  'correctly?')

    keys = ['NEO4J_HOME', 'NEO4J_INSTANCE', 'JAVA_OPTS', 'CLASSPATH']

    result = {}

    for lne in output.splitlines():
        for key in keys:
            sep = key + ':'
            if lne.startswith(sep):
                _, value = lne.split(sep)
                value = value.strip()
                result[key] = value

    return result


def temp_neo4j_instance(uri):
    """ Start or return an existing instance of the neo4j graph database server
    using the given URI.
    """

    # split the uri
    match = re.match(r"temp://?(?P<port>\d*)(?P<temp_dir>.*)", uri)
    port = match.group("port") or '7475'

    # return http uri if this temporary database already exists,
    # using the port as the unique identifer
    if port in _temporary_databases:
        return "http://localhost:{}/db/data/".format(port)

    neo4j_info = get_neo4j_info()

    temp_dir = match.group("temp_dir") or tempfile.mkdtemp()
    temp_data_dir = os.path.join(temp_dir, 'data')
    if not os.path.exists(temp_data_dir):
        os.makedirs(temp_data_dir)

    # otherwise, start a new neo4j process.
    # define the subprocess command and the required classpath
    cmd = ['java', '-cp']
    cmd.append(neo4j_info['CLASSPATH'])

    # neo requires a physical config file which we temporarily create to
    # specify config overrides
    config_options = {
        'org.neo4j.server.webserver.port': port,
        'org.neo4j.server.database.location': temp_data_dir
    }

    props_filepath = os.path.join(temp_dir, 'server.properties')
    with open(props_filepath, 'w') as props_file:
        for key, value in config_options.iteritems():
            props_file.write("{}={}\n".format(key, value))

    # required startup args
    args = [
        '-server', '-XX:+DisableExplicitGC'
    ]

    # additional startup options for the command line
    startup_options = {
        'neo4j.home': neo4j_info['NEO4J_HOME'],
        'neo4j.instance': neo4j_info['NEO4J_INSTANCE'],
        'org.neo4j.server.properties': props_filepath,
        'file.encoding': 'UTF-8',
    }
    for key, value in startup_options.iteritems():
        args.append("-D{}={}".format(key, value))

    # finally add the class to run
    args.append('org.neo4j.server.Bootstrapper')
    cmd.extend(args)

    # start the subprocess
    neo4j_process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
    _temporary_databases[port] = neo4j_process

    # terminate subprocess at exit
    def terminate():  # pragma: no cover
        try:
            neo4j_process.terminate()
        except OSError:
            pass  # already terminated
    atexit.register(terminate)

    # the startup process is async so we monitor the http interface to know
    # when to allow the test runner to continue
    timeout_time = time.time() + TIMEOUT
    url = "http://localhost:{}/db/data/".format(port)

    while time.time() < timeout_time:
        try:
            req = requests.get(url)
            if "neo4j_version" in req.text:
                logger.debug('neo4j server started on {}'.format(url))
                return url  # return REST API url
        except requests.ConnectionError:
            time.sleep(0.2)

    logger.critical(
        'Unable to start Neo4j: timeout after %s '
        'seconds. See logs in %s.', TIMEOUT, temp_data_dir)

    raise TempConnectionError(url)


def get_connection(uri):
    if uri.startswith('temp://'):
        uri = temp_neo4j_instance(uri)
    graph_db = neo4j.GraphDatabaseService(uri)
    return graph_db
